#include <sys/types.h>
#include <sys/stat.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <stdint.h>
#include <fcntl.h>

#include "common.h"
#include "macros.h"
#include "io.h"
#include "signal-event.h"
#include "misc.h"
#include "timeout.h"
#include "buffer.h"
#include "hal-proto.h"
#include "hal-peer.h"
#include "hal-cmd-parser.h"

#define info(fmt, arg...) fprintf(stderr, "info: %s: "fmt CRLF, \
				__FUNCTION__, ##arg);

#define error(eval, fmt, arg...)			\
do {							\
	fprintf(stderr, "error: %s: "fmt CRLF,		\
				__FUNCTION__, ##arg);	\
	exit(eval);					\
} while (0)

static struct io_context *io;

/* Handler for request commands. */
static void connect_request(void *opaque)
{
	info("connect request: OK");
}

static void disconnect_request(void *opaque)
{
	info("disconnect request: OK");
}

static void cycle_begin_request(void *opaque)
{
	info("cycle begin request: OK");
}

static void cycle_run_request(void *opaque)
{
	info("cycle run request: OK");
}

static void cycle_end_request(void *opaque)
{
	info("cycle end request: OK");
}

static void write_prepare_request(void *opaque)
{
	info("write prepare request: OK");
}

static void write_end_request(void *opaque)
{
	info("write end request: OK");
}

static void write_inputs_request(const char *image, unsigned int off, void *opaque)
{
	info("write inputs request: OK");
}

static void write_outputs_request(const char *image, unsigned int off, void *opaque)
{
	info("write outputs request: OK");
}

static void read_prepare_request(void *opaque)
{
	info("read prepare request: OK");
}

static void read_end_request(void *opaque)
{
	info("read end request: OK");
}

static void read_inputs_request(unsigned int len, unsigned int off,
				void *opaque)
{
	info("read inputs request: OK");
}

static void read_outputs_request(unsigned int len, unsigned int off,
				 void *opaque)
{
	info("read outputs request: OK");
}

/* Handlers for reply commands. */
static void connect_reply(unsigned int inputs_sz, unsigned int outputs_sz,
			  char side, const char *error, void *opaque)
{
	if (error != NULL) {
		info("connect reply: %s", error);
		return;
	}

	info("connect reply: OK");
}

static void disconnect_reply(const char *error, void *opaque)
{
	if (error != NULL) {
		info("disconnect reply: %s", error);
		return;
	}

	info("disconnect reply: OK");
	io_loop_stop(io);
}

static void cycle_begin_reply(const char *error, void *opaque)
{
	if (error != NULL) {
		info("cycle begin reply: %s", error);
		return;
	}

	info("cycle run reply: OK");
}

static void cycle_run_reply(const char *error, void *opaque)
{
	if (error != NULL) {
		info("cycle run reply: %s", error);
		return;
	}

	info("cycle run reply: OK");
}

static void cycle_end_reply(const char *error, void *opaque)
{
	if (error != NULL) {
		info("cycle end reply: %s", error);
		return;
	}

	info("cycle end reply: OK");
}

static void write_inputs_reply(const char *error, void *opaque)
{
	if (error != NULL) {
		info("write inputs reply: %s", error);
		return;
	}

	info("write inputs reply: OK");
}

static void write_outputs_reply(const char *error, void *opaque)
{
	if (error != NULL) {
		info("write outputs reply: %s", error);
		return;
	}

	info("write outputs reply: OK");
}

static void write_prepare_reply(const char *error, void *opaque)
{
	if (error != NULL) {
		info("write prepare reply: %s", error);
		return;
	}

	info("write prepare reply: OK");
}

static void write_end_reply(const char *error, void *opaque)
{
	if (error != NULL) {
		info("write end reply: %s", error);
		return;
	}

	info("write end reply: OK");
}

static void read_prepare_reply(const char *error, void *opaque)
{
	if (error != NULL) {
		info("read prepare reply: %s", error);
		return;
	}

	info("read prepare reply: OK");
}

static void read_end_reply(const char *error, void *opaque)
{
	if (error != NULL) {
		info("read end reply: %s", error);
		return;
	}

	info("read end reply: OK");
}

static void read_inputs_reply(const char *inputs, unsigned int off,
			      const char *error, void *opaque)
{
	if (error != NULL) {
		info("read inputs reply: %s", error);
		return;
	}

	info("read inputs reply: OK");
}

static void read_outputs_reply(const char *outputs, unsigned int off,
			       const char *error, void *opaque)
{
	if (error != NULL) {
		info("read outputs reply: %s", error);
		return;
	}

	info("read outputs reply: OK");
}

static void peer_event_handler(struct hal_peer_base *peer,
			       hal_peer_event_t event)
{
	switch (event) {
	case HAL_PEER_EVENT_ERROR:
	case HAL_PEER_EVENT_TIMEOUT:
		info("got error or timeout event");
		io_loop_stop(peer->io);
		break;
	case HAL_PEER_EVENT_READ:
	case HAL_PEER_EVENT_WRITE:
		break;
	}
}

static void init_handlers(struct hal_cmd_callbacks *callbacks)
{
	memset(callbacks, 0, sizeof(struct hal_cmd_callbacks));

	callbacks->connect_request_callback = connect_request;
	callbacks->disconnect_request_callback = disconnect_request;
	callbacks->cycle_begin_request_callback = cycle_begin_request;
	callbacks->cycle_run_request_callback = cycle_run_request;
	callbacks->cycle_end_request_callback = cycle_end_request;
	callbacks->write_prepare_request_callback = write_prepare_request;
	callbacks->write_end_request_callback = write_end_request;
	callbacks->write_inputs_request_callback = write_inputs_request;
	callbacks->write_outputs_request_callback = write_outputs_request;
	callbacks->read_prepare_request_callback = read_prepare_request;
	callbacks->read_end_request_callback = read_end_request;
	callbacks->read_inputs_request_callback = read_inputs_request;
	callbacks->read_outputs_request_callback = read_outputs_request;

	callbacks->connect_reply_callback = connect_reply;
	callbacks->disconnect_reply_callback = disconnect_reply;
	callbacks->cycle_begin_reply_callback = cycle_begin_reply;
	callbacks->cycle_run_reply_callback = cycle_run_reply;
	callbacks->cycle_end_reply_callback = cycle_end_reply;
	callbacks->write_prepare_reply_callback = write_prepare_reply;
	callbacks->write_end_reply_callback = write_end_reply;
	callbacks->write_inputs_reply_callback = write_inputs_reply;
	callbacks->write_outputs_reply_callback = write_outputs_reply;
	callbacks->read_prepare_reply_callback = read_prepare_reply;
	callbacks->read_end_reply_callback = read_end_reply;
	callbacks->read_inputs_reply_callback = read_inputs_reply;
	callbacks->read_outputs_reply_callback = read_outputs_reply;
}

static void handle_signals(int signo,  siginfo_t *si, void *opaque)
{
	switch (signo) {
	case SIGINT:
	case SIGTERM:
		info("exiting by signal %d", signo);
		io_loop_stop(io);
		break;
	}
}

int main(int argc, char **argv)
{
	struct hal_peer_base *p;
	struct hal_cmd_callbacks callbacks;
	int rfd, wfd;
	ret_t ret;

	if (argc < 2)
		error(1, "not enough arguments: write fifo read fifo");

	/* Open pipes and get file descriptors. */
	ret = pipe_open(argv[1], O_RDWR, &wfd);
	if (ret != ret_ok)
		error(1, "can't open write pipe");

	ret = pipe_open(argv[2], O_RDWR, &rfd);
	if (ret != ret_ok)
		error(1, "can't open read pipe");

	/* Set signal handlers. */
	sigevent_ignore(SIGPIPE);
	sigevent_add(SIGINT, handle_signals, NULL, 1);
	sigevent_add(SIGTERM, handle_signals, NULL, 1);
	sigevent_init();

	/* Init loop facilities. */
	timeout_init();
	io = io_new();

	/* Create a single peer. */
	p = hal_peer_base_new();
	hal_peer_base_init(p, io, wfd, rfd, peer_event_handler);

	/* Prepare a parser and command handlers. */
	init_handlers(&callbacks);
	hal_cmd_parser_register_callbacks(&callbacks);

	/* Wait and parse commands. */
	io_loop_run(io);	

	/* Free resources. */
	hal_peer_base_deinit(p);
	hal_peer_base_free(&p);

	io_free(&io);
	timeout_deinit();
	sigevent_deinit();

	return 0;
}

